# -*- coding: utf-8 -*-

"""Core classes."""

import logging

from enum import Enum, auto
from random import choice, sample
from typing import Iterable, List, Set, Tuple

from .utils import pairwise

LOGGER = logging.getLogger(__name__)


class Action(Enum):
    """Possible actions."""

    TAKE = auto()
    PASS = auto()


ACTIONS = tuple(Action)


class Game:
    """A game of No Thanks!"""

    NUM_PLAYERS_MIN = 3
    NUM_PLAYERS_MAX = 7
    CARD_MIN = 3
    CARD_MAX = 35
    NUM_CARDS_DISCARD = 9

    players: Tuple["Player", ...]
    draw_pile: List[int]
    current_player: int
    tokens_on_card: int

    def __init__(self: "Game", players: Iterable["Player"]) -> None:
        self.players = tuple(players)
        assert self.NUM_PLAYERS_MIN <= len(self.players) <= self.NUM_PLAYERS_MAX
        self.reset()

    def reset(self: "Game") -> None:
        """Reset the game."""

        self.current_player = 0
        self.tokens_on_card = 0
        self.players = tuple(sample(population=self.players, k=len(self.players)))

        tokens_per_player = min(55 // len(self.players), 11)
        for player in self.players:
            player.reset(game=self, tokens=tokens_per_player)

        self.draw_pile = sample(
            population=range(self.CARD_MIN, self.CARD_MAX + 1),
            k=self.CARD_MAX - self.NUM_CARDS_DISCARD,
        )

        LOGGER.info("Draw pile: %s", ", ".join(map(str, self.draw_pile)))

    @property
    def finished(self: "Game") -> bool:
        """Has this game finished?"""
        return not self.draw_pile

    @property
    def sort_players(self: "Game") -> Tuple["Player", ...]:
        """Sort players by their score."""
        attr = "score" if self.finished else "score_cards"
        return tuple(sorted(self.players, key=lambda player: -getattr(player, attr)))

    def play(self: "Game") -> Tuple["Player", ...]:
        """Play the game."""

        LOGGER.info("Starting a game with %d players", len(self.players))

        while not self.finished:
            player = self.players[self.current_player]
            LOGGER.info(
                "Card: %d; token(s): %d; card(s) remaining: %d, active player: %s",
                self.draw_pile[0],
                self.tokens_on_card,
                len(self.draw_pile) - 1,
                player,
            )

            action = player.action()
            assert action is Action.TAKE or player.tokens > 0
            LOGGER.info("%s chose action: %s", player, action)

            if action is Action.TAKE:
                player.cards.add(self.draw_pile.pop(0))
                player.tokens += self.tokens_on_card
                LOGGER.info(
                    "%s now has %d token(s) and runs: %s",
                    player,
                    player.tokens,
                    player.runs,
                )
                self.tokens_on_card = 0

            else:
                player.tokens -= 1
                self.tokens_on_card += 1
                self.current_player += 1
                while self.current_player >= len(self.players):
                    self.current_player -= len(self.players)

        LOGGER.info("Game has finished, final results:")

        results = self.sort_players

        for pos, player in enumerate(results):
            LOGGER.info(
                "#%d %s with %d points: %d token(s) and %s",
                pos + 1,
                player,
                player.score,
                player.tokens,
                player.runs,
            )

        return results


def _make_runs(cards: Iterable[int]) -> Iterable[Iterable[int]]:
    cards = sorted(cards)
    if not cards:
        return
    run = [cards[0]]
    for prev, succ in pairwise(cards):
        if succ == prev + 1:
            run.append(succ)
        else:
            yield run
            run = [succ]
    yield run


class Player:
    """A player."""

    name: str
    game: Game
    tokens: int
    cards: Set[int]

    def __init__(self: "Player", name: str) -> None:
        self.name = name

    def __str__(self: "Player") -> str:
        return f"Player <{self.name}>"

    def reset(self: "Player", game: Game, tokens: int) -> None:
        """Reset the player."""
        self.game = game
        self.tokens = tokens
        self.cards = set()

    def action(self: "Player") -> Action:
        """Choose an action."""
        return Action.TAKE if self.tokens <= 0 else choice(ACTIONS)

    @property
    def runs(self: "Player") -> Tuple[Tuple[int, ...], ...]:
        """The sequencial runs formed by this player's cards."""
        return tuple(tuple(run) for run in _make_runs(self.cards))

    @property
    def score_cards(self: "Player") -> int:
        """Minus points incurred from cards."""
        return sum(-run[0] for run in self.runs)

    @property
    def score(self: "Player") -> int:
        """Total score."""
        return self.score_cards + self.tokens



# -*- coding: utf-8 -*-

"""Players."""

import logging

from random import random

import inquirer

from .core import Action, Player

LOGGER = logging.getLogger(__name__)


class Human(Player):
    """Human interactive input."""

    def action(self: "Human") -> Action:
        """Prompt for an action."""

        if self.tokens <= 0:
            return Action.TAKE

        action_tags = [
            inquirer.questions.TaggedValue(label=str(a), value=a)
            for a in reversed(Action)
        ]
        question = inquirer.List(
            name="action",
            message=f"Choose an action [{self.tokens} token(s) in hand; runs: {self.runs}]",
            choices=action_tags,
            carousel=True,
        )
        answer = inquirer.prompt([question])

        assert answer
        assert isinstance(answer.get("action"), Action)

        return answer["action"]


class Heuristic(Player):
    """Use heuristics to choose actions."""

    def action(self: "Heuristic") -> Action:
        """Choose an action based on heuristics."""
        proba = self.take_proba()
        LOGGER.info("Probability to take: %.3f", proba)
        return Action.TAKE if self.tokens <= 0 or random() <= proba else Action.PASS

    def take_proba(self: "Heuristic") -> float:
        """Probability to play TAKE."""

        card = self.game.draw_pile[0]
        tokens_on_card = self.game.tokens_on_card
        value = card - tokens_on_card - 1

        if self.tokens <= 0 or value <= 0:
            return 1

        proba = 0.05 if value >= 20 else 1 - value / 20

        proba *= 1 + 1 / self.tokens

        if (card - 1) in self.cards or (card + 1) in self.cards:
            proba *= 2

        return max(min(proba, 1), 0)



# -*- coding: utf-8 -*-

"""CLI entry point."""

import logging
import sys

from typing import List

from .core import Game, Player
from .players import Heuristic, Human


def main() -> None:
    """CLI entry point."""

    logging.basicConfig(
        stream=sys.stderr,
        level=logging.INFO,
        format="%(levelname)-4.4s [%(name)s:%(lineno)s] %(message)s",
    )

    num_players = 5
    players: List[Player] = [Human(name="You")]
    players += [Heuristic(name=f"AI #{i + 1}") for i in range(num_players - 1)]
    game = Game(players=players)
    game.play()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Utility functions."""

from itertools import tee
from typing import Iterable, Tuple, TypeVar

TYPE = TypeVar("TYPE")


def pairwise(iterable: Iterable[TYPE]) -> Iterable[Tuple[TYPE, TYPE]]:
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    it1, it2 = tee(iterable)
    next(it2, None)
    return zip(it1, it2)
